#! /usr/bin/python3

from .cmd_reader import reader
from .service import Service
from .builder import ZoneBuilder
from .process import (
    launch_background_process, 
    stop_background_process
)

def main():
    cmd, args, kwargs = reader()
    if not any([cmd, args, kwargs]):
        service = Service()
        service.listen()
    
    if cmd == ["config"]:
        zb = ZoneBuilder(*args, **kwargs)
        zb.main()

    elif cmd == ["start"]:
        launch_background_process(*args, **kwargs)

    elif cmd == ["stop"]:
        stop_background_process(*args, **kwargs)

    
if __name__ == "__main__":
    main()



import sys

from .help_menu import HelpMenu

def simple_isnan(item):
    try:
        float(item)
        return False
    except ValueError:
        return True

def _get_help(cmd=""):
    help_menu = HelpMenu()
    help_menu.get(cmd)
    sys.exit()

def reader():
    cmd = sys.argv[1:2]
    args = sys.argv[2::]
    kwargs = {}

    if cmd in ("-h", "--help"):
        return _get_help()
    elif any(h in args for h in ("-h", "--help")):
        return _get_help(cmd)
        
    kwarg_keys = reversed([
        index for index, item in enumerate(args) 
        if (item.startswith('-') and simple_isnan(item))
    ])
    for k in kwarg_keys:
        try:
            value = args.pop(k+1)
        except IndexError:
            value = True
        key = args.pop(k)
        kwargs[key] = value

    return cmd, args, kwargs


import os
import signal
import sys

from .service import Service


HERE = os.path.abspath(os.path.dirname(__file__))


def launch_background_process(*args, **kwargs):
    print("STARTING...")
    try:
        pid = os.fork()
        if pid > 0:
            sys.exit(0)
    except OSError as e:
        print(sys.stderr, f"FORK FAILED: {e.errno} ({e.strerror})")
        sys.exit(1)

    pid = str(os.getpid())
    
    with open(os.path.join(HERE, '.pid'), 'w') as f:
        f.write(pid)

    print(f"STARTED PID: {pid}")
    service = Service()
    service.listen()
    

def stop_background_process(*args, **kwargs):
    with open(os.path.join(HERE, '.pid'), 'r') as f:
        pid = int(f.read())
    print(f"STOPPING...")
    os.kill(pid, signal.SIGTERM)
    print(f"STOPPED PID: {pid}")




import json
import os
import sys
import threading

from Xlib import X
from Xlib.display import Display
from Xlib.error import ConnectionClosedError

from .snap import active_window


HERE = os.path.abspath(os.path.dirname(__file__))


class ZoneBuilder:

    def __init__(self, *args, **kwargs) -> None:
        self.zones = []
        self.display = Display()
        self.screen = self.display.screen()
        self.running = True

        if count := kwargs.get('-n'):
            self.add(count)

    def _main_action(self):
        return input(
            'specify action:\n$ '
        )

    def _get_count(self):
        return input(
            'specify count:\n$ '
        )

    def main(self):
        action = self._main_action()

        if action == "save":
            self.save

        elif action == "exit":
            self.terminate() 

        else:
            print('action not understood. Please try again.\n')

        self.main()

    @property
    def save(self):
        results = []
        for zone in self.zones:
            # for some reason this works better with
            # a new display object on each iteration
            window = active_window(Display(), zone.id)  
            pg = window.query_tree().parent.query_tree().parent.get_geometry()
            results.append({
                "x": pg.x,
                "y": pg.y,
                "width": pg.width,
                "height": pg.height
            })
        self._write(results)
        self.terminate()

    def add(self, count):
        for _ in range(int(count)):
            window = self.screen.root.create_window(
                10, 10, 500, 250, 1,
                self.screen.root_depth,
                background_pixel=750000,
                event_mask=X.ExposureMask | X.KeyPressMask,
            )
            window.map()
            self.zones.append(window)
        thread = threading.Thread(target=self.loop)
        thread.daemon = True
        thread.start()

    def loop(self):
        while True:
            try:
                e = self.display.next_event()
            except ConnectionClosedError:
                # raised when X button is pressed
                #
                # TODO: figure out way to allow exit of single window without
                # having to kill the entire application (currently get a
                # "socket_error" if you try to add a window after exiting 
                # an already made window)
                #
                # for now, just kill everything
                print()
                os._exit(1)

    def _read(self, path = os.path.join(HERE, "zones.json")):
        with open(path, 'r') as f:
            return json.loads(f.read())

    def _write(self, results, path = os.path.join(HERE, "zones.json")):
        with open(path, 'w') as f:
            f.write(json.dumps(results, indent=2, sort_keys=True))

    def terminate(self):
        sys.exit()


import json
import os.path

HERE = os.path.abspath(os.path.dirname(__file__))


class Zone:
    def __init__(self, x, y, width, height) -> None:
        self.x = x
        self.y = y
        self.width = width
        self.height = height

    def check(self, x, y):
        if (
            self.x <= x <= self.x + self.width
        ) and (
            self.y <= y <= self.y + self.height
        ):
            return True
        return False


class ZoneProfile:
    def __init__(self, zones) -> None:
        self.zones = zones

    def find_zone(self, x, y, shift=None):
        for index, item in enumerate(self.zones):
            if item.check(x, y):
                if not shift:
                    return item
                elif shift == "LEFT":
                    return self.zones[(index-1)%len(self.zones)]
                elif shift == "RIGHT":
                    return self.zones[(index+1)%len(self.zones)]
        return None

    @staticmethod
    def from_file(path = os.path.join(HERE, "zones.json")):
        if os.path.isfile(path):
            with open(path, 'r') as f:
                data = json.loads(f.read())
            return ZoneProfile([Zone(**obj) for obj in data])
        return ZoneProfile([])



from Xlib import X, XK
from Xlib.ext import record
from Xlib.display import Display
from Xlib.protocol import rq

from .snap import snap_window, shift_window


class Service:
    def __init__(self) -> None:
        self.keybindings = {
            XK.XK_s: False,
            XK.XK_Alt_L: False
        }

        self.display = Display()
        self.root = self.display.screen().root
        
        self.context = self.display.record_create_context(0, [record.AllClients], [{
            'core_requests': (0, 0),
            'core_replies': (0, 0),
            'ext_requests': (0, 0, 0, 0),
            'ext_replies': (0, 0, 0, 0),
            'delivered_events': (0, 0),
            'device_events': (X.KeyReleaseMask, X.ButtonReleaseMask),
            'errors': (0, 0),
            'client_started': False,
            'client_died': False,
        }])

        self.display.record_enable_context(self.context, self.handler)
        self.display.record_free_context(self.context)
        
    def handler(self, reply):
        data = reply.data
        while len(data):

            event, data = rq.EventField(None).parse_binary_value(data, self.display.display, None, None)
                
            if event.type == X.KeyPress or X.KeyRelease:
                keysym = self.display.keycode_to_keysym(event.detail, 0)
                if keysym in self.keybindings:
                    self.keybindings[keysym] = (
                        True if event.type == X.KeyPress else False
                    )

            if all([value == True for value in self.keybindings.values()]):

                if event.type == X.ButtonRelease:
                    snap_window(event.root_x, event.root_y)

                elif event.type == X.KeyPress:
                    keysym = self.display.keycode_to_keysym(event.detail, 0)
                    if keysym == XK.XK_Left:
                        shift_window('LEFT')

                    elif keysym == XK.XK_Right:
                        shift_window('RIGHT')

    def listen(self):
        while True:
            self.root.display.next_event()